#include "mainapp.h"

#include "browserserver.h"
#include "CefBrowser.h"
#include <thrift/transport/TSocket.h>
#include <thrift/protocol/TBinaryProtocol.h>
#include <thrift/transport/TBufferTransports.h>
#include <thrift/server/TThreadPoolServer.h>
#include <thrift/transport/TServerSocket.h>
#include <thrift/transport/TBufferTransports.h>

using namespace ::apache::thrift;
using namespace ::apache::thrift::protocol;
using namespace ::apache::thrift::transport;
using namespace ::apache::thrift::server;
using namespace ::apache::thrift::concurrency;
using namespace ::cefbrowser;
using namespace ::common;

#include <utility>
#include "browserclient.h"
#include "v8handler.h"
#include "tools.h"
#include "keycodes.h"
#include "database.h"

using namespace ::apache::thrift;
using namespace ::apache::thrift::protocol;
using namespace ::apache::thrift::transport;
using namespace ::apache::thrift::server;
using namespace ::apache::thrift::concurrency;
using namespace ::common;

scoped_refptr<CefBrowser> currentBrowser;
TThreadPoolServer *thriftServer;

void startServer(BrowserParameter bParam) {
    BrowserServer::setBrowserParameter(bParam);

    const int workerCount = 4;
    std::shared_ptr<ThreadManager> threadManager = ThreadManager::newSimpleThreadManager(workerCount);
    threadManager->threadFactory(std::make_shared<ThreadFactory>());
    threadManager->start();

    std::string listenIp = bParam.bindAll ? "0.0.0.0" : bParam.browserIp;

    thriftServer = new TThreadPoolServer(
            std::make_shared<CefBrowserProcessorFactory>(std::make_shared<BrowserCloneFactory>()),
            std::make_shared<TServerSocket>(listenIp, bParam.browserPort),
            std::make_shared<TBufferedTransportFactory>(),
            std::make_shared<TBinaryProtocolFactory>(),
            threadManager);

    thriftServer->serve();
}

bool stopCheckVdrRegular = false;

void checkVdrRegular() {
    while (!stopCheckVdrRegular) {
        if (currentBrowser != nullptr) {
            std::string url = currentBrowser->GetMainFrame()->GetURL().ToString();
            if (!url.empty() && url != BLANK_PAGE) {
                auto c = dynamic_cast<BrowserClient *>(currentBrowser->GetHost()->GetClient().get());
                if (!c->IsVdrWebActive()) {
                    currentBrowser->GetMainFrame()->LoadURL(BLANK_PAGE);
                }
            }
        }

        std::this_thread::sleep_for(std::chrono::milliseconds(500));
    }
}

// BrowserApp
BrowserApp::BrowserApp(BrowserParameter bparam) : bParameter(bparam) {
    CefMessageRouterConfig config;
    config.js_query_function = "cefQuery";
    config.js_cancel_function = "cefQueryCancel";
}

CefRefPtr<CefBrowserProcessHandler> BrowserApp::GetBrowserProcessHandler() {
    return this;
}

CefRefPtr<CefRenderProcessHandler> BrowserApp::GetRenderProcessHandler() {
    return this;
}

void BrowserApp::OnBeforeCommandLineProcessing(const CefString& process_type, CefRefPtr<CefCommandLine> command_line) {
    command_line->AppendSwitchWithValue("autoplay-policy", "no-user-gesture-required");
}

void BrowserApp::OnContextInitialized() {
    LOG_CURRENT_THREAD();

    DEBUG("BrowserApp::OnContextInitialized()");

    CefBrowserSettings browserSettings;
    browserSettings.windowless_frame_rate = 25;

    CefWindowInfo window_info;
    window_info.SetAsWindowless(0);
    window_info.shared_texture_enabled = false;

    CefRefPtr<CefDictionaryValue> extra_info = CefDictionaryValue::Create();
    extra_info->SetString("browserIp", bParameter.browserIp);
    extra_info->SetInt("browserPort", bParameter.browserPort);
    extra_info->SetString("transcoderIp", bParameter.transcoderIp);
    extra_info->SetInt("transcoderPort", bParameter.transcoderPort);
    extra_info->SetString("vdrIp", bParameter.vdrIp);
    extra_info->SetInt("vdrPort", bParameter.vdrPort);
    extra_info->SetInt("LogLevel", logger->level());
    extra_info->SetInt("osdqoi", (int)bParameter.osdqoi);
    extra_info->SetBool("dirtyRecs", bParameter.use_dirty_recs);

    CefRefPtr<BrowserClient> client = new BrowserClient(false, bParameter);
    currentBrowser = CefBrowserHost::CreateBrowserSync(window_info, client, "", browserSettings, extra_info, nullptr);

    // read user agents if configured
    if (!bParameter.user_agent_path.empty()) {
        database->readUserAgents(bParameter.user_agent_path);
    }

    INFO("Start Server on {}:{} with static path {}", bParameter.browserIp, bParameter.browserPort, bParameter.static_path);

    std::thread t1(startServer, bParameter);
    t1.detach();

    std::thread t2(checkVdrRegular);
    t2.detach();
}

void BrowserApp::OnWebKitInitialized() {
}

void BrowserApp::OnBrowserCreated(CefRefPtr<CefBrowser> browser, CefRefPtr<CefDictionaryValue> extra_info) {
    bParameter.browserIp = extra_info->GetString("browserIp");
    bParameter.browserPort = extra_info->GetInt("browserPort");
    bParameter.transcoderIp = extra_info->GetString("transcoderIp");
    bParameter.transcoderPort = extra_info->GetInt("transcoderPort");
    bParameter.vdrIp = extra_info->GetString("vdrIp");
    bParameter.vdrPort = extra_info->GetInt("vdrPort");
    bParameter.osdqoi = (image_type_enum)extra_info->GetInt("osdqoi");
    bParameter.use_dirty_recs = extra_info->GetBool("dirtyRecs");

    int logLevel = extra_info->GetInt("LogLevel");
    logger->set_level(static_cast<spdlog::level::level_enum>(logLevel));
}

void BrowserApp::OnContextCreated(CefRefPtr<CefBrowser> browser, CefRefPtr<CefFrame> frame, CefRefPtr<CefV8Context> context) {
    TRACE("BrowserApp::OnContextCreated");

    // register all native JS functions
    CefRefPtr<CefV8Value> object = context->GetGlobal();
    handler = new V8Handler(bParameter);

    CefRefPtr<CefV8Value> streamVideo = CefV8Value::CreateFunction("StreamVideo", handler);
    object->SetValue("cefStreamVideo", streamVideo, V8_PROPERTY_ATTRIBUTE_NONE);

    CefRefPtr<CefV8Value> stopVideo = CefV8Value::CreateFunction("StopVideo", handler);
    object->SetValue("cefStopVideo", stopVideo, V8_PROPERTY_ATTRIBUTE_NONE);

    CefRefPtr<CefV8Value> pauseVideo = CefV8Value::CreateFunction("PauseVideo", handler);
    object->SetValue("cefPauseVideo", pauseVideo, V8_PROPERTY_ATTRIBUTE_NONE);

    CefRefPtr<CefV8Value> resumeVideo = CefV8Value::CreateFunction("ResumeVideo", handler);
    object->SetValue("cefResumeVideo", resumeVideo, V8_PROPERTY_ATTRIBUTE_NONE);

    CefRefPtr<CefV8Value> seekVideo = CefV8Value::CreateFunction("SeekVideo", handler);
    object->SetValue("cefSeekVideo", seekVideo, V8_PROPERTY_ATTRIBUTE_NONE);

    CefRefPtr<CefV8Value> redButton = CefV8Value::CreateFunction("RedButton", handler);
    object->SetValue("cefRedButton", redButton, V8_PROPERTY_ATTRIBUTE_NONE);

    CefRefPtr<CefV8Value> loadUrl = CefV8Value::CreateFunction("LoadUrl", handler);
    object->SetValue("cefLoadUrl", loadUrl, V8_PROPERTY_ATTRIBUTE_NONE);

    CefRefPtr<CefV8Value> videoSize = CefV8Value::CreateFunction("VideoSize", handler);
    object->SetValue("cefVideoSize", videoSize, V8_PROPERTY_ATTRIBUTE_NONE);

    CefRefPtr<CefV8Value> videoFull = CefV8Value::CreateFunction("VideoFullscreen", handler);
    object->SetValue("cefVideoFullscreen", videoFull, V8_PROPERTY_ATTRIBUTE_NONE);

    CefRefPtr<CefV8Value> startApp = CefV8Value::CreateFunction("StartApp", handler);
    object->SetValue("cefStartApp", startApp, V8_PROPERTY_ATTRIBUTE_NONE);

    CefRefPtr<CefV8Value> audioInfo = CefV8Value::CreateFunction("AudioInfo", handler);
    object->SetValue("cefAudioInfo", audioInfo, V8_PROPERTY_ATTRIBUTE_NONE);

    CefRefPtr<CefV8Value> audioTrack = CefV8Value::CreateFunction("SelectAudioTrack", handler);
    object->SetValue("cefSelectAudioTrack", audioTrack, V8_PROPERTY_ATTRIBUTE_NONE);
}

void BrowserApp::OnContextReleased(CefRefPtr<CefBrowser> browser, CefRefPtr<CefFrame> frame, CefRefPtr<CefV8Context> context) {
    TRACE("BrowserApp::OnContextReleased");
    handler = nullptr;
}

bool BrowserApp::OnProcessMessageReceived(CefRefPtr<CefBrowser> browser, CefRefPtr<CefFrame> frame, CefProcessId source_process, CefRefPtr<CefProcessMessage> message) {
    TRACE("BrowserApp::OnProcessMessageReceived(Renderer) {}", message->GetName().ToString());

    return false;
}

void BrowserApp::shutdown() {
    if (thriftServer) {
        thriftServer->stop();
    }

    stopCheckVdrRegular = true;

    currentBrowser->Release();
    currentBrowser = nullptr;
}

